-- Create pgstatspack_snap procedure.
--
-- By frits.hoogland@interaccess.nl
-- Based on Glenn.Fawcett@Sun.com's snap procedure
--

CREATE OR REPLACE FUNCTION pgstatspack_snap ( description varchar(256) ) RETURNS bigint AS $$
DECLARE
  now_dts TIMESTAMP;
  spid BIGINT;
  version_major int;
  version_minor int;
BEGIN
SELECT current_timestamp INTO now_dts; 
  SELECT nextval('pgstatspackid') INTO spid;
  INSERT INTO pgstatspack_snap values (spid, now_dts, description);

INSERT INTO pgstatspack_database
SELECT
  spid               as snapid,
  d.datid            as datid,
  d.datname         as datname,
  d.numbackends     as numbackends,
  d.xact_commit         as xact_commit,
  d.xact_rollback    as xact_rollback,
  d.blks_read        as blks_read,
  d.blks_hit        as blks_hit
FROM
  pg_stat_database d
;

INSERT INTO pgstatspack_tables
SELECT
  spid               as snapid,
  t.schemaname ||'.'|| t.relname          as table_name,
  t.seq_scan         as seq_scan,
  t.seq_tup_read     as seq_tup_read,
  t.idx_scan         as idx_scan,
  t.idx_tup_fetch    as idx_tup_fetch,
  t.n_tup_ins        as n_tup_ins,
  t.n_tup_upd        as n_tup_upd,
  t.n_tup_del        as n_tup_del,
  it.heap_blks_read  as heap_blks_read,
  it.heap_blks_hit   as heap_blks_hit,
  it.idx_blks_read   as idx_blks_read,
  it.idx_blks_hit    as idx_blks_hit,
  it.toast_blks_read as toast_blks_read,
  it.toast_blks_hit  as toast_blks_hit,
  it.tidx_blks_read  as tidx_blks_read,
  it.tidx_blks_hit   as tidx_blks_hit,
  pg_relation_size(t.relid)+pg_relation_size(s.relid) as tbl_size,
  sum(pg_relation_size(i.indexrelid)) as idx_size
FROM
  pg_statio_all_tables it,
  pg_stat_all_tables t
  JOIN pg_class c on t.relid=c.oid
  JOIN pg_stat_sys_tables s on c.reltoastrelid=s.relid 
  JOIN pg_index i on i.indrelid=t.relid
WHERE
  (t.relid = it.relid)
GROUP BY
  t.schemaname,t.relname,t.seq_scan,t.seq_tup_read,t.idx_scan,t.idx_tup_fetch,t.n_tup_ins,t.n_tup_upd,t.n_tup_del,it.heap_blks_read,it.heap_blks_hit,it.idx_blks_read,it.idx_blks_hit,it.toast_blks_read,it.toast_blks_hit,it.tidx_blks_read,it.tidx_blks_hit,t.relid,s.relid
;

INSERT INTO pgstatspack_indexes
SELECT
  spid               as snapid,
  i.schemaname ||'.'|| i.indexrelname     as index_name,
  i.schemaname ||'.'|| i.relname          as table_name,
  i.idx_scan         as idx_scan,
  i.idx_tup_read     as idx_tup_read,
  i.idx_tup_fetch    as idx_tup_fetch,
  ii.idx_blks_read   as idx_blks_read,
  ii.idx_blks_hit    as idx_blks_hit
FROM
  pg_stat_all_indexes i,
  pg_statio_all_indexes ii
WHERE
  (i.indexrelid = ii.indexrelid)
;

INSERT INTO pgstatspack_sequences
SELECT
  spid               as snapid,
  s.schemaname       as sequence_schema,
  s.relname          as sequence_name,
  s.blks_read        as seq_blks_read,
  s.blks_hit         as seq_blks_hit
FROM
  pg_statio_all_sequences s
;

INSERT INTO pgstatspack_settings
SELECT
  spid			as snapid,
  s.name		as name,
  s.setting		as setting,
  s.source		as source
FROM
  pg_settings s
WHERE
  (s.source != 'default')
;

select cast(substring(version(), 'PostgreSQL ([0-9]*).') as int) into version_major;
select cast(substring(version(), 'PostgreSQL [0-9]*.([0-9]*).') as int) into version_minor;

IF ((version_major = 8 AND version_minor >= 4 ) OR version_major > 8 ) THEN
 BEGIN
  INSERT INTO pgstatspack_statements
  SELECT
    spid as snapid,
    s.userid as userid,
    s.dbid as dbid,
    s.query as query,
    s.calls as calls,
    s.total_time as total_time,
    s.rows as rows
  FROM pg_stat_statements s
  ORDER BY total_time
  LIMIT 20;
 EXCEPTION WHEN object_not_in_prerequisite_state THEN
  raise warning '%', SQLERRM;

 select pg_stat_statements_reset();
 END; 
END IF;

RETURN spid;
END;
$$ LANGUAGE plpgsql;

